from typing import Protocol, Sequence

from .backend import Backend
from .utils import cur_ts, logger
from .worker import WORKER_TTL_IN_SEC


class Service(Protocol):
    def __call__(self) -> float:
        """Call the service.

        It should returns the time interval to wait for next call.
        """
        ...


class WorkerThreadProtocol(Protocol):
    def get_id(self) -> str:
        ...

    def is_alive(self) -> bool:
        ...


class RefreshWorkerLifetime(Service):
    def __init__(self,
                 backend: Backend,
                 workers: Sequence[WorkerThreadProtocol],
                 **kwargs):
        self.backend = backend
        self.workers = list(workers)

    def __call__(self) -> float:
        self.backend.set_worker_ttl(
            {w.get_id() for w in self.workers if w.is_alive()},
            cur_ts() + WORKER_TTL_IN_SEC)
        return 1.0


class PurgeDeadWorkers(Service):
    def __init__(self, backend: Backend, **kwargs):
        self.backend = backend
        self.lock = backend.get_lock('purge_workers')

    def __call__(self) -> float:
        if self.lock.acquire():
            try:
                now = cur_ts()
                to_purge = {
                    wid for (wid, ttl) in self.backend.get_workers()
                    if ttl < now
                }
                if to_purge:
                    self.backend.purge_workers(to_purge)
            finally:
                self.lock.release()
        return 10.0


class RestoreAbandonedTasks(Service):
    def __init__(self, backend: Backend, **kwargs):
        self.backend = backend
        self.lock = backend.get_lock('restore_abandoned_tasks')

    def __call__(self) -> float:
        if self.lock.acquire():
            try:
                stage_info_list = self.backend.get_stage_info(50)
                now = cur_ts()
                active_worker_ids = {
                    wid for (wid, ttl) in self.backend.get_workers()
                    if ttl >= now
                }
                for stage_info in stage_info_list:
                    if stage_info.worker_id not in active_worker_ids:
                        logger.info(f'restore task `{stage_info.task_id}`')
                        self.backend.restore(stage_info)
            finally:
                self.lock.release()
        return 5.0


class DiscardDisposableTasks(Service):
    def __init__(self, backend: Backend, **kwargs):
        self.backend = backend
        self.lock = backend.get_lock('discard_disposable_tasks')

    def __call__(self) -> float:
        if self.lock.acquire():
            try:
                task_ids = self.backend.get_disposable_task_ids(limit=100)
                if task_ids:
                    logger.info(f'discard {len(task_ids)} tasks ({", ".join(task_ids)})')
                    self.backend.discard_tasks(*task_ids)
                if len(task_ids) == 100:
                    return 0
            finally:
                self.lock.release()
        return 60.0



import json
from typing import Protocol, Generator, Literal, TypedDict, Optional, Union


class Shutdown(TypedDict):
    name: Literal['shutdown']
    groups: Optional[list[str]]


class Pause(TypedDict):
    name: Literal['pause']
    groups: Optional[list[str]]


class Resume(TypedDict):
    name: Literal['resume']
    groups: Optional[list[str]]


ControlEvent = Union[Shutdown, Pause, Resume]


class EventBridge(Protocol):
    def receive_events(self) -> Generator[ControlEvent, None, None]:
        ...

    def send_event(self, event: ControlEvent):
        ...


def encode_control_event(event: ControlEvent) -> bytes:
    return json.dumps(event).encode('utf-8')


def decode_control_event(encoded: bytes) -> Optional[ControlEvent]:
    try:
        event: ControlEvent = json.loads(encoded.decode('utf-8'))
        assert isinstance(event['name'], str)
        assert event['groups'] is None or isinstance(event['groups'], list)
    except Exception:
        return None
    return event



from traceback import format_exc
from typing import Any
from uuid import uuid4

from .backend import Backend
from .result import Result, prevent_to_wait_result, ResultGetPrevented
from .task import Task, TaskHandler, DiscardTask
from .utils import logger


REASON_TO_PREVENT_WAIT_RESULT =\
    'You should avoid to wait results of other tasks because it may causes '\
    'deadlock. If you can assure that it is safe you can pass '\
    '`avoid_assertion=True` for the `get` method.'

WORKER_TTL_IN_SEC = 10


class Worker:
    def __init__(self,
                 group: str,
                 backend: Backend,
                 handler: TaskHandler):
        self.id = f'wk_{group}_{uuid4().hex}'
        self.group = group
        self.backend = backend
        self.handler = handler

    def __call__(self) -> bool:
        """Pop a task from the queue and handle the task.

        If there is a task to handle then it returns whether if task finished
        successfully otherwise returns False.
        """

        try:
            task = self.backend.assign_task(self.group, self.id)
            if task is None:
                return False
            else:
                self._handle_task(task)
                return True
        except Exception:
            logger.exception(f'[{self.id}] unexpected exception')
            return False

    def _handle_task(self, task: Task):
        logger.info(f'[{self.id}] handle task ({task.id}: {task.name})')
        try:
            with prevent_to_wait_result(REASON_TO_PREVENT_WAIT_RESULT):
                ret = self.handler.encode_result(
                    task, self.handler.handle(task))
        except DiscardTask:
            self._discard_task(task)
        except ResultGetPrevented:
            self._handle_prevented(task)
        except Exception as e:
            self._handle_error(task, e)
        else:
            self.backend.succeed(task, ret)

    def _discard_task(self, task: Task):
        logger.info(f'[{self.id}] task was discarded ({task.id}: {task.name})')
        self.backend.discard_tasks(task.id)

    def _handle_prevented(self, task: Task):
        logger.exception(f'[{self.id}] Result.get was prevented during '
                         f'handling a task ({task.name}).')
        self.backend.fail(task, format_exc())

    def _handle_error(self, task: Task, exception: Exception):
        try:
            with prevent_to_wait_result(REASON_TO_PREVENT_WAIT_RESULT):
                interval = self.handler.get_retry_interval(task, exception)
            if interval is None:
                logger.info(f'[{self.id}] task was failed '
                            f'({task.id}: {task.name})')
                self.backend.fail(task, format_exc())
                return
        except DiscardTask:
            self._discard_task(task)
        except ResultGetPrevented:
            self._handle_prevented(task)
        except Exception:
            handler_name = self.handler.__class__.__qualname__
            logger.exception(
                f'[{self.id}] {handler_name}.get_retry_interval raises error '
                f'so we assume that we should retry ({task.id}: {task.name}')
            self._retry(task, task.retry_count + 1)
        else:
            self._retry(task, interval)

    def _retry(self, task: Task, interval: float):
        clone = task.clone_for_retry(interval)
        logger.info(f'[{self.id}] retry n{clone.retry_count} '
                    f'({task.id}: {task.name})')
        self.backend.put_tasks(clone)


class EagerWorker(Worker):
    def __init__(self, backend: Backend, handler: TaskHandler):
        super().__init__('_', backend, handler)

    def __call__(self) -> bool:
        raise AssertionError('Can not call EagerWorker')

    def handle_task(self, task: Task) -> Result[Any]:
        self._handle_task(task)
        return Result(self.backend, self.handler, task.id)

    def _handle_error(self, task: Task, exception: Exception):
        self.backend.fail(task, format_exc())



import json
from typing import Generator, Optional

from redis.client import Redis, Pipeline
from redis.exceptions import LockError
from redis.lock import Lock as _RedisLock

from ..backend import Backend, Lock, NotFound, NoResult, Failed
from ..event import EventBridge, ControlEvent, encode_control_event, decode_control_event
from ..kit import Kit
from ..stage import StageInfo
from ..task import Task, TaskHandler
from ..utils import cur_ts


class RedisEventBridge(EventBridge):
    def __init__(self, redis: Redis):
        self.channel = 'taskkit.redis_bridge.channel'
        self.redis = redis

    def receive_events(self) -> Generator[ControlEvent, None, None]:
        pubsub = self.redis.pubsub()
        pubsub.subscribe(self.channel)
        while True:
            for m in pubsub.listen():
                if not isinstance(m['data'], bytes):
                    continue
                if (event := decode_control_event(m['data'])):
                    yield event

    def send_event(self, event: ControlEvent):
        self.redis.publish(self.channel, encode_control_event(event))


class RedisLock(Lock):
    def __init__(self, lock: _RedisLock):
        self.lock = lock

    def acquire(self) -> bool:
        return self.lock.acquire(blocking=False)

    def release(self):
        try:
            self.lock.release()
        except LockError:
            pass


_KEY_PREFIX = 'taskkit.backend'


class RedisBackend(Backend):
    def __init__(self, redis: Redis):
        self.redis = redis

    def workers_key(self) -> str:
        return f'{_KEY_PREFIX}.workers'

    def queue_key(self, group: str) -> str:
        return f'{_KEY_PREFIX}.{group}.queue'

    def data_store_key(self) -> str:
        return f'{_KEY_PREFIX}.data_store'

    def task_info_key(self) -> str:
        return f'{_KEY_PREFIX}.task_info'

    def stage_queue_key(self) -> str:
        return f'{_KEY_PREFIX}.stage_queue'

    def stage_info_key(self) -> str:
        return f'{_KEY_PREFIX}.stage_info'

    def done_queue_key(self) -> str:
        return f'{_KEY_PREFIX}.done_queue'

    def disposable_queue_key(self) -> str:
        return f'{_KEY_PREFIX}.disposable_queue'

    def result_key(self) -> str:
        return f'{_KEY_PREFIX}.result'

    def error_message_key(self) -> str:
        return f'{_KEY_PREFIX}.error_message'

    def scheduler_data_key(self, name: str) -> str:
        return f'{_KEY_PREFIX}.scheduler.{name}'

    def lock_key_prefix(self, target: str) -> str:
        return f'{_KEY_PREFIX}.locks.{target}'

    def set_worker_ttl(self, worker_ids: set[str], expires_at: float):
        if not worker_ids:
            return
        self.redis.zadd(self.workers_key(), {
            worker_id: expires_at
            for worker_id in worker_ids
        })

    def get_workers(self) -> list[tuple[str, float]]:
        return [
            (worker_id.decode(), ttl) for worker_id, ttl in
            self.redis.zrange(self.workers_key(), 0, -1, withscores=True)
        ]

    def purge_workers(self, worker_ids: set[str]):
        if worker_ids:
            self.redis.zrem(self.workers_key(), *worker_ids)

    def put_tasks(self, *tasks: Task):
        if not tasks:
            return
        with self.redis.pipeline() as pipe:
            self._put_tasks(pipe, *tasks)
            pipe.execute()

    def _put_tasks(self, pipe: Pipeline, *_tasks: Task):
        if not _tasks:
            return
        for group, tasks in Task.group_tasks(*_tasks).items():
            task_ids = [t.id for t in tasks]
            pipe.zrem(self.stage_queue_key(), *task_ids)
            pipe.hdel(self.stage_info_key(), *task_ids)
            self._put_task_data(pipe, *_tasks)
            pipe.zadd(self.queue_key(group), {t.id: t.due for t in tasks})

    def _put_task_data(self, pipe: Pipeline, *tasks: Task):
        if not tasks:
            return
        pipe.hset(self.data_store_key(), mapping={
            t.id: t.data for t in tasks
        })
        pipe.hset(self.task_info_key(), mapping={
            t.id: self._encode_info(t) for t in tasks
        })

    def get_queued_tasks(self, group: str, limit: int) -> list[Task]:
        task_ids = list(
            self.redis.zrange(self.queue_key(group), 0, limit))
        tasks = self.lookup_tasks(*task_ids)
        return [t for tid in task_ids if (t := tasks.get(tid))]

    def lookup_tasks(self, *task_ids: str) -> dict[str, Optional[Task]]:
        if not task_ids:
            return dict()

        info = self.redis.hmget(self.task_info_key(), *task_ids)
        data = self.redis.hmget(self.data_store_key(), *task_ids)

        return {
            tid: Task(**json.loads(i), data=d)
            if (i is not None and d is not None) else None
            for tid, i, d in zip(task_ids, info, data)
        }

    def assign_task(self, group: str, worker_id: str) -> Optional[Task]:
        def _pop(pipe):
            ret: list[tuple[bytes, float]] =\
                pipe.zrange(self.queue_key(group), 0, 0, withscores=True)
            if not ret:
                return None
            task_id, score = ret[0]
            if score > cur_ts():
                return None
            info = json.loads(
                pipe.hget(self.task_info_key(), task_id).decode())
            task = Task(**info, data=pipe.hget(self.data_store_key(), task_id))

            pipe.multi()
            pipe.zrem(self.queue_key(group), task_id)
            pipe.zadd(self.stage_queue_key(), {task_id: task.due})
            pipe.hset(self.stage_info_key(),
                      task.id, StageInfo.init(worker_id, task).to_json())
            return task

        return self.redis.transaction(
            _pop, self.queue_key(group), value_from_callable=True)

    def discard_tasks(self, *task_ids: str):
        if not task_ids:
            return

        with self.redis.pipeline() as pipe:
            for group, _task_ids in Task.group_task_ids(*task_ids).items():
                pipe.zrem(self.queue_key(group), *_task_ids)
            pipe.zrem(self.stage_queue_key(), *task_ids)
            pipe.hdel(self.stage_info_key(), *task_ids)
            pipe.hdel(self.data_store_key(), *task_ids)
            pipe.hdel(self.task_info_key(), *task_ids)
            pipe.zrem(self.done_queue_key(), *task_ids)
            pipe.zrem(self.disposable_queue_key(), *task_ids)
            pipe.hdel(self.result_key(), *task_ids)
            pipe.hdel(self.error_message_key(), *task_ids)
            pipe.execute()

    def succeed(self, task: Task, result: bytes):
        group = task.group
        with self.redis.pipeline() as pipe:
            pipe.zrem(self.queue_key(group), task.id)
            pipe.zrem(self.stage_queue_key(), task.id)
            pipe.hdel(self.stage_info_key(), task.id)
            done = cur_ts()
            pipe.zadd(self.done_queue_key(), {task.id: done})
            pipe.zadd(self.disposable_queue_key(),
                      {task.id: done + task.ttl})
            pipe.hset(self.result_key(), task.id, result)
            self._put_task_data(pipe, task)
            pipe.execute()

    def fail(self, task: Task, error_message: str):
        group = task.group
        with self.redis.pipeline() as pipe:
            pipe.zrem(self.queue_key(group), task.id)
            pipe.zrem(self.stage_queue_key(), task.id)
            pipe.hdel(self.stage_info_key(), task.id)
            done = cur_ts()
            pipe.zadd(self.done_queue_key(), {task.id: done})
            pipe.zadd(self.disposable_queue_key(),
                      {task.id: done + task.ttl})
            pipe.hset(self.error_message_key(), task.id,
                      error_message.encode('utf-8'))
            self._put_task_data(pipe, task)
            pipe.execute()

    def get_result(self, task_id: str) -> tuple[Task, bytes]:
        task = self.lookup_tasks(task_id).get(task_id)
        if task is None:
            raise NotFound

        ret = self.redis.hget(self.result_key(), task_id)
        if ret is not None:
            return (task, ret)

        message = self.redis.hget(self.error_message_key(), task_id)
        if message is not None:
            raise Failed(task, message.decode('utf-8'))

        raise NoResult(task)

    def get_done_task_ids(self,
                          since: Optional[float],
                          until: Optional[float],
                          limit: int) -> list[str]:
        return [
            v.decode() for v in self.redis.zrangebyscore(
                self.done_queue_key(),
                since or 0,
                until or cur_ts(),
                start=0,
                num=limit)
        ]

    def get_disposable_task_ids(self, limit: int) -> list[str]:
        return [
            v.decode() for v in self.redis.zrangebyscore(
                self.disposable_queue_key(), 0, cur_ts(), start=0, num=limit)
        ]

    def get_stage_info(self, limit: int) -> list[StageInfo]:
        task_ids = list(
            self.redis.zrange(self.stage_queue_key(), 0, limit))
        if not task_ids:
            return []
        info_data = self.redis.hmget(self.stage_info_key(), *task_ids)
        return [StageInfo.from_json(d) for d in info_data if d is not None]

    def restore(self, info: StageInfo):
        task = self.lookup_tasks(info.task_id).get(info.task_id)
        if task is None:
            return
        with self.redis.pipeline() as pipe:
            pipe.zrem(self.stage_queue_key(), task.id)
            pipe.hdel(self.stage_info_key(), task.id)
            pipe.zadd(self.queue_key(task.group), {task.id: task.due})
            pipe.execute()

    def get_lock(self, target: str) -> Lock:
        return RedisLock(
            self.redis.lock(self.lock_key_prefix(target), timeout=10))

    def persist_scheduler_state_and_put_tasks(self,
                                              name: str,
                                              data: bytes,
                                              *tasks: Task):
        with self.redis.pipeline() as pipe:
            pipe.set(self.scheduler_data_key(name), data)
            self._put_tasks(pipe, *tasks)
            pipe.execute()

    def get_scheduler_state(self, name: str) -> Optional[bytes]:
        return self.redis.get(self.scheduler_data_key(name))

    def destroy_all(self):
        keys = self.redis.keys(f'{_KEY_PREFIX}.*')
        if keys:
            self.redis.delete(*keys)

    def _encode_info(self, task: Task) -> str:
        data = task.to_dict()
        del data['data']
        return json.dumps(data)


def make_kit(redis: Redis,
             handler: TaskHandler) -> Kit:
    backend = RedisBackend(redis)
    bridge = RedisEventBridge(redis)
    return Kit(backend, bridge, handler)



from datetime import datetime, tzinfo
from logging import getLogger
from time import time
from typing import overload, Optional, Union


logger = getLogger('taskkit')


_LOCAL_TZ = datetime.now().astimezone().tzinfo


def local_tz() -> tzinfo:
    assert _LOCAL_TZ is not None
    return _LOCAL_TZ


cur_ts = time


@overload
def as_ts(value: Union[datetime, float]) -> float:
    ...


@overload
def as_ts(value: Union[datetime, float, None]) -> Optional[float]:
    ...


def as_ts(value: Union[datetime, float, None]) -> Optional[float]:
    if value is None:
        return None
    if isinstance(value, datetime):
        return value.timestamp()
    return value


def from_ts(ts: float, tzinfo: tzinfo) -> datetime:
    return datetime.fromtimestamp(ts, tzinfo)



from django.db import models


class TaskkitControlEvent(models.Model):
    class Meta:
        db_table = 'taskkit_control_event'
        app_label = 'taskkit'

    sent: models.FloatField = models.FloatField(db_index=True)
    data: models.BinaryField = models.BinaryField()


class TaskkitLock(models.Model):
    class Meta:
        db_table = 'taskkit_lock'
        app_label = 'taskkit'

    id: models.CharField = models.CharField(primary_key=True, max_length=255)


class TaskkitWorker(models.Model):
    class Meta:
        db_table = 'taskkit_worker'
        app_label = 'taskkit'

    id: models.CharField = models.CharField(primary_key=True, max_length=255)
    expires: models.FloatField = models.FloatField(db_index=True)


class TaskkitTask(models.Model):
    class Meta:
        db_table = 'taskkit_task'
        app_label = 'taskkit'
        index_together = (
            ('began', 'group', 'due'),
            ('done', 'began'),
            ('name', 'created'),
        )

    id: models.CharField = models.CharField(primary_key=True, max_length=255)
    group: models.CharField = models.CharField(max_length=40)
    name: models.CharField = models.CharField(max_length=255)
    data: models.BinaryField = models.BinaryField()
    due: models.FloatField = models.FloatField()
    created: models.FloatField = models.FloatField()
    scheduled: models.FloatField = models.FloatField(null=True)
    retry_count: models.PositiveIntegerField = models.PositiveIntegerField()
    ttl: models.FloatField = models.FloatField()

    assignee_worker_id: models.CharField = models.CharField(null=True, max_length=255)
    began: models.FloatField = models.FloatField(null=True)

    result: models.BinaryField = models.BinaryField(null=True)
    error_message: models.TextField = models.TextField(null=True)
    done: models.FloatField = models.FloatField(null=True)
    disposable: models.FloatField = models.FloatField(null=True, db_index=True)


class TaskkitSchedulerState(models.Model):
    class Meta:
        db_table = 'taskkit_scheduler_state'
        app_label = 'taskkit'

    id: models.CharField = models.CharField(primary_key=True, max_length=255)
    data: models.BinaryField = models.BinaryField()



from contextlib import contextmanager
from contextvars import ContextVar
from enum import Enum
from time import time, sleep
from typing import Generic, TypeVar, overload, cast, Optional, Union

from .backend import Backend, NoResult
from .task import TaskHandler


T = TypeVar('T')


class _Sentinel(Enum):
    obj = object()


class ResultGetTimedOut(Exception):
    pass


class ResultGetPrevented(Exception):
    pass


_prevent_to_wait_result: ContextVar[tuple[bool, str]] =\
    ContextVar('prevent_to_wait_result', default=(False, ''))


@contextmanager
def prevent_to_wait_result(reason: str):
    token = _prevent_to_wait_result.set((True, reason))
    try:
        yield
    finally:
        _prevent_to_wait_result.reset(token)


class Result(Generic[T]):
    @overload
    def __init__(self,
                 backend: Backend,
                 handler: TaskHandler,
                 task_id: str):
        ...

    @overload
    def __init__(self,
                 backend: Backend,
                 handler: TaskHandler,
                 task_id: str,
                 *,
                 result: T):
        ...

    def __init__(self,
                 backend: Backend,
                 handler: TaskHandler,
                 task_id: str,
                 result: Union[T, _Sentinel] = _Sentinel.obj):
        self.backend = backend
        self.handler = handler
        self.task_id = task_id
        self._result = result

    def get(self,
            timeout: Optional[float] = None,
            avoid_assertion: bool = False) -> T:
        if not avoid_assertion:
            prevent, reason = _prevent_to_wait_result.get()
            if prevent:
                raise ResultGetPrevented(reason)

        if not isinstance(self._result, _Sentinel):
            return cast(T, self._result)

        try:
            return self.__get()
        except NoResult:
            pass

        i = 0
        t = time()
        while timeout is None or (time() - t) < timeout:
            try:
                return self.__get()
            except NoResult:
                sleep(min(i * 0.1, 1))
        raise ResultGetTimedOut()

    def __get(self) -> T:
        task, encoded = self.backend.get_result(self.task_id)
        ret = self.handler.decode_result(task, encoded)
        self._result = ret
        return ret



from typing import Protocol, Optional

from .task import Task
from .stage import StageInfo


class Lock(Protocol):
    def acquire(self) -> bool:
        """try to acquire the lock"""
        ...

    def release(self):
        """release the lock"""
        ...


class NotFound(Exception):
    pass


class NoResult(Exception):
    def __init__(self, task: Task):
        self.task = task


class Failed(Exception):
    def __init__(self, task: Task, message: str):
        self.task = task
        self.message = message


class Backend(Protocol):
    def set_worker_ttl(self, worker_ids: set[str], expires_at: float):
        """Update the workers lifetime"""
        ...

    def get_workers(self) -> list[tuple[str, float]]:
        """Get list of pairs of worker id and expiration timestamp"""
        ...

    def purge_workers(self, worker_ids: set[str]):
        """purge given workers"""
        ...

    def put_tasks(self, *tasks: Task):
        """Put the tasks to the queue

        If there is same task on the stage, it should be removed from the
        stage first because it will be called when the task should retry.
        """
        ...

    def get_queued_tasks(self, group: str, limit: int) -> list[Task]:
        """Get tasks in the queue"""
        ...

    def lookup_tasks(self, *task_ids: str) -> dict[str, Optional[Task]]:
        """Lookup tasks by given ids"""
        ...

    def assign_task(self, group: str, worker_id: str) -> Optional[Task]:
        """Take a task from the queue and assign it for the worker and put on
        the stage."""
        ...

    def discard_tasks(self, *task_ids: str):
        """Discard the tasks"""
        ...

    def succeed(self, task: Task, result: bytes):
        """Save the result and remove the task from the stage. It should
        save the task data if the task is missing in the backend."""
        ...

    def fail(self, task: Task, error_message: str):
        """Save the error message and remove the task from the stage. It should
        save the task data if the task is missing in the backend."""
        ...

    def get_result(self, task_id: str) -> tuple[Task, bytes]:
        """It returns a tuple of the task and the result for given task_id or
        raises suitable exception.

        It raises
        - NotFound if the task was not found
        - Failed if the task was failed
        - NoResult if the task is not done
        """
        ...

    def get_done_task_ids(self,
                          since: Optional[float],
                          until: Optional[float],
                          limit: int) -> list[str]:
        """Get done task ids matching given timestamp range

        It may contain failed task ids.
        """
        ...

    def get_disposable_task_ids(self, limit: int) -> list[str]:
        """Get disposable task ids

        Tasks will be treated as disposable after `ttl` seconds from
        the time the task was succeeded or failed.
        """
        ...

    def get_stage_info(self, limit: int) -> list[StageInfo]:
        """Get stage info"""
        ...

    def restore(self, info: StageInfo):
        """Remove the task from the stage and put it back to the queue"""
        ...

    def get_lock(self, target: str) -> Lock:
        """Get a lock object for the target"""
        ...

    def persist_scheduler_state_and_put_tasks(self,
                                              name: str,
                                              data: bytes,
                                              *tasks: Task):
        """Persist given data for the scheduler and put tasks"""
        ...

    def get_scheduler_state(self, name: str) -> Optional[bytes]:
        """Get data for the scheduler"""
        ...

    def destroy_all(self):
        """Destroy all data which associated with the backend. It is needed
        for unit tests."""
        ...



import signal
from contextlib import contextmanager
from types import FrameType
from typing import Optional


class SignalCaptured(BaseException):
    def __init__(self, signal: int, frame: Optional[FrameType]):
        self.signal = signal
        self.frame = frame

    @classmethod
    def handler(cls, signal: int, frame: Optional[FrameType]):
        raise cls(signal, frame)


@contextmanager
def capture_signals(*signals: signal.Signals):
    originals = [signal.getsignal(s) for s in signals]

    def handler(*args, **kwargs):
        SignalCaptured.handler(*args, **kwargs)

    for s in signals:
        signal.signal(s, handler)

    try:
        yield
    finally:
        for s, org in zip(signals, originals):
            signal.signal(s, org)



import json
from dataclasses import dataclass, asdict
from datetime import datetime, tzinfo
from typing import Protocol, Literal, Optional, Union

from .backend import Backend
from .services import Service
from .task import Task
from .utils import cur_ts, from_ts, as_ts, logger


# in seconds
SCHEDULE_POINT_INTERVAL = 5


class Schedule(Protocol):
    def get_timezone(self) -> Optional[tzinfo]:
        ...

    def __call__(self,
                 schedule_points: list[datetime],
                 last_scheduled_at: Optional[datetime],
                 /) -> list[datetime]:
        ...


class DuplicationPolicy(Protocol):
    def __call__(self,
                 schedule_points: list[datetime],
                 last_scheduled_at: Optional[datetime],
                 /) -> list[datetime]:
        ...


class OnlyEarliest(DuplicationPolicy):
    def __call__(self,
                 schedule_points: list[datetime],
                 last_scheduled_at: Optional[datetime],
                 /) -> list[datetime]:
        return schedule_points[:1]


class OnlyLatest(DuplicationPolicy):
    def __call__(self,
                 schedule_points: list[datetime],
                 last_scheduled_at: Optional[datetime],
                 /) -> list[datetime]:
        return schedule_points[-1:]


class RegularSchedule(Schedule):
    _seconds = set(range(0, 60, SCHEDULE_POINT_INTERVAL))
    _minutes = set(range(60))
    _hours = set(range(24))
    _days = set(range(1, 32))
    _weekdays = set(range(7))
    _months = set(range(1, 13))

    def __init__(self,
                 seconds: Union[Literal['*'], set[int], int, None] = 0,
                 minutes: Union[Literal['*'], set[int], int, None] = None,
                 hours: Union[Literal['*'], set[int], int, None] = None,
                 days: Union[Literal['*'], set[int], int, None] = None,
                 weekdays: Union[Literal['*'], set[int], int, None] = None,
                 months: Union[Literal['*'], set[int], int, None] = None,
                 tzinfo: Optional[tzinfo] = None,
                 duplication_policy: DuplicationPolicy = OnlyLatest()):
        self.seconds = self._ensure('seconds', seconds, self._seconds)
        self.minutes = self._ensure('minutes', minutes, self._minutes)
        self.hours = self._ensure('hours', hours, self._hours)
        self.days = self._ensure('days', days, self._days)
        self.weekdays = self._ensure('weekdays', weekdays, self._weekdays)
        self.months = self._ensure('months', months, self._months)
        self.tzinfo = tzinfo
        self.duplication_policy = duplication_policy

    def get_timezone(self) -> Optional[tzinfo]:
        if self.tzinfo is None:
            return None
        return self.tzinfo

    @staticmethod
    def _ensure(key: str,
                value: Union[Literal['*'], set[int], int, None],
                all_valid_values: set[int]) -> set[int]:
        if value is None or value == '*':
            return all_valid_values
        if isinstance(value, int):
            value = {value}
        assert all(t in all_valid_values for t in value),\
            f'All values of `{key}` must be in {all_valid_values}.'\
            f' The values are: `{value}`.'
        return value

    def __call__(self,
                 schedule_points: list[datetime],
                 last_scheduled_at: Optional[datetime],
                 /) -> list[datetime]:
        return self.duplication_policy(
            [p for p in schedule_points if self._filter(p)],
            last_scheduled_at)

    def _filter(self, schedule_point: datetime) -> bool:
        if schedule_point.second not in self.seconds:
            return False
        if schedule_point.minute not in self.minutes:
            return False
        if schedule_point.hour not in self.hours:
            return False
        if schedule_point.day not in self.days:
            return False
        if schedule_point.weekday() not in self.weekdays:
            return False
        if schedule_point.month not in self.months:
            return False
        return True


@dataclass(frozen=True)
class ScheduleEntry:
    key: str
    schedule: Schedule
    group: str
    name: str
    data: bytes


@dataclass(frozen=True)
class SchedulerState:
    last_run_at: float

    # key: Entry.key
    # value: last scheudled timestamp for the entry
    last_scheduled_at: dict[str, float]


class Scheduler(Service):
    def __init__(self,
                 name: str,
                 backend: Backend,
                 entries: list[ScheduleEntry],
                 tzinfo: tzinfo):
        assert len(entries) == len({e.key for e in entries}),\
            'All entries must have unique keys'
        self.name = name
        self.backend = backend
        self.entries = entries
        self.lock = backend.get_lock(f'scheduler.{name}')
        self.tzinfo = tzinfo

    def __call__(self) -> float:
        """It schedules entries and returns time interval indicating when
        should this method be called next time."""

        start = self._round(cur_ts())

        if self.entries and self.lock.acquire():
            try:
                self._schedule_entries()
            finally:
                self.lock.release()

        return max(0, (start + SCHEDULE_POINT_INTERVAL) - cur_ts())

    def _schedule_entries(self):
        state = self._get_state()
        schedule_points = self._list_schedule_points(state)
        if not schedule_points:
            return

        ls_at = state.last_scheduled_at if state else {}
        new_state = SchedulerState(
            last_run_at=schedule_points[-1],
            last_scheduled_at={},
        )
        tasks: list[Task] = []
        for e in self.entries:
            last = ls_at.get(e.key)
            tz = e.schedule.get_timezone() or self.tzinfo
            if last is not None:
                new_state.last_scheduled_at[e.key] = last
                last = from_ts(last, tz)
            for sp in e.schedule(
                    [from_ts(sp, tz) for sp in schedule_points], last):
                new_state.last_scheduled_at[e.key] = as_ts(sp)
                task = Task.init(group=e.group, name=e.name, data=e.data,
                                 due=sp, scheduled=sp)
                tasks.append(task)
                logger.info(f'schedule task at {sp} ({task.id}: {task.name})')

        self.backend.persist_scheduler_state_and_put_tasks(
            self.name,
            self._encode_state(new_state),
            *tasks)

    def _encode_state(self, state: SchedulerState) -> bytes:
        return json.dumps(asdict(state)).encode()

    def _get_state(self) -> Optional[SchedulerState]:
        data = self.backend.get_scheduler_state(self.name)
        if data is None:
            return None

        return SchedulerState(**json.loads(data.decode()))

    def _list_schedule_points(
            self, state: Optional[SchedulerState]) -> list[float]:
        at = self._round(cur_ts())
        if state is None:
            return [at]
        if state.last_run_at >= at:
            return []

        n = int((at - state.last_run_at) // SCHEDULE_POINT_INTERVAL)
        return [
            state.last_run_at + (i * SCHEDULE_POINT_INTERVAL)
            for i in range(1, n + 1)
        ]

    def _round(self, ts: float) -> float:
        return ts // SCHEDULE_POINT_INTERVAL * SCHEDULE_POINT_INTERVAL



import json
from dataclasses import dataclass, asdict
from typing import Any

from .task import Task
from .utils import cur_ts


@dataclass
class StageInfo:
    worker_id: str
    task_id: str
    began: float

    @classmethod
    def init(cls,
             worker_id: str,
             task: Task) -> 'StageInfo':
        return cls(worker_id=worker_id,
                   task_id=task.id,
                   began=cur_ts())

    def to_dict(self) -> dict[str, Any]:
        return asdict(self)

    def to_json(self) -> str:
        return json.dumps(self.to_dict())

    @classmethod
    def from_json(cls, json_data: str) -> 'StageInfo':
        return cls(**json.loads(json_data))



import time
from datetime import datetime, tzinfo
from typing import Any, Callable, Mapping, Sequence, TypedDict, Optional, Union

from .backend import Backend
from .event import EventBridge, Shutdown, Pause, Resume
from .process import TaskkitProcess
from .result import Result
from .scheduler import ScheduleEntry, Schedule
from .signal import SignalCaptured, capture_signals, signal
from .task import Task, TaskHandler, DEFAULT_TASK_TTL
from .utils import local_tz
from .worker import EagerWorker


class ScheduleEntryDict(TypedDict):
    key: str
    schedule: Schedule
    group: str
    name: str
    data: Any


ScheduleEntryCompat = Union[ScheduleEntry, ScheduleEntryDict]
ScheduleEntriesCompat = Sequence[ScheduleEntryCompat]
ScheduleEntriesCompatMapping = Mapping[str, ScheduleEntriesCompat]


class Kit:
    def __init__(self,
                 backend: Backend,
                 bridge: EventBridge,
                 handler: TaskHandler):
        self.backend = backend
        self.bridge = bridge
        self.handler = handler
        self.eager_worker = EagerWorker(backend, handler)

    def start(self,
              num_processes: int,
              num_worker_threads_per_group: dict[str, int],
              schedule_entries: ScheduleEntriesCompatMapping = {},
              tzinfo: Optional[tzinfo] = None,
              should_restart: Callable[[TaskkitProcess], bool] = lambda _: False):

        schedule_entries = self._ensure_schedule_entries(schedule_entries)

        def _start():
            return self._start_process(
                num_worker_threads_per_group=num_worker_threads_per_group,
                schedule_entries=schedule_entries,
                tzinfo=tzinfo,
                daemon=True)

        processes = [_start() for _ in range(num_processes)]

        try:
            with capture_signals(signal.SIGTERM):
                while True:
                    for i, p in enumerate(list(processes)):
                        if not p.is_alive() or should_restart(p):
                            if p.is_active():
                                p.terminate()
                            p.join()
                            processes[i] = _start()
                    time.sleep(1)
        except (KeyboardInterrupt, SystemExit, SignalCaptured) as e:
            for p in processes:
                if p.is_active():
                    p.terminate()
            for p in processes:
                p.join()
            if isinstance(e, SignalCaptured):
                raise SystemExit from e
            else:
                raise

    def start_processes(self,
                        num_processes: int,
                        num_worker_threads_per_group: dict[str, int],
                        schedule_entries: ScheduleEntriesCompatMapping = {},
                        tzinfo: Optional[tzinfo] = None,
                        daemon: bool = True) -> list[TaskkitProcess]:
        schedule_entries = self._ensure_schedule_entries(schedule_entries)
        return [
            self._start_process(num_worker_threads_per_group,
                                schedule_entries,
                                tzinfo,
                                daemon)
            for _ in range(num_processes)
        ]

    def _start_process(self,
                       num_worker_threads_per_group: dict[str, int],
                       schedule_entries: dict[str, list[ScheduleEntry]] = {},
                       tzinfo: Optional[tzinfo] = None,
                       daemon: bool = True) -> TaskkitProcess:
        p = TaskkitProcess(
            num_worker_threads_per_group=num_worker_threads_per_group,
            backend=self.backend,
            bridge=self.bridge,
            handler=self.handler,
            schedule_entries=schedule_entries,
            tzinfo=tzinfo or local_tz(),
            daemon=daemon)
        p.start()
        return p

    def _ensure_schedule_entries(self, entries: ScheduleEntriesCompatMapping)\
            -> dict[str, list[ScheduleEntry]]:

        return {
            k: [self._ensure_schedule_entry(e) for e in v]
            for k, v in entries.items()
        }

    def _ensure_schedule_entry(self, entry: ScheduleEntryCompat)\
            -> ScheduleEntry:
        if isinstance(entry, dict):
            return ScheduleEntry(
                key=entry['key'],
                schedule=entry['schedule'],
                group=entry['group'],
                name=entry['name'],
                data=self.handler.encode_data(
                    entry['group'], entry['name'], entry['data']),
            )
        return entry

    def initiate_task(self,
                      group: str,
                      name: str,
                      data: Any,
                      due: Optional[datetime] = None,
                      ttl: float = DEFAULT_TASK_TTL,
                      eager: bool = False) -> Result[Any]:
        encoded = self.handler.encode_data(group, name, data)
        task = Task.init(group, name=name, data=encoded, due=due, ttl=ttl)
        if eager:
            return self.eager_worker.handle_task(task)
        self.backend.put_tasks(task)
        return Result(self.backend, self.handler, task.id)

    def send_shutdown_event(self, groups: Optional[set[str]] = None):
        event: Shutdown = {
            'name': 'shutdown',
            'groups': None if groups is None else list(groups),
        }
        self.bridge.send_event(event)

    def send_pause_event(self, groups: Optional[set[str]] = None):
        event: Pause = {
            'name': 'pause',
            'groups': None if groups is None else list(groups),
        }
        self.bridge.send_event(event)

    def send_resume_event(self, groups: Optional[set[str]] = None):
        event: Resume = {
            'name': 'resume',
            'groups': None if groups is None else list(groups),
        }
        self.bridge.send_event(event)



import sys
import time
from typing import Generator, Optional

from django.db.transaction import atomic, Atomic
from django.db.utils import OperationalError, IntegrityError

from ..backend import Backend, Lock, NotFound, NoResult, Failed
from ..contrib.django.models import TaskkitControlEvent, TaskkitLock, TaskkitWorker, TaskkitTask, TaskkitSchedulerState
from ..event import EventBridge, ControlEvent, encode_control_event, decode_control_event
from ..kit import Kit
from ..stage import StageInfo
from ..task import Task, TaskHandler
from ..utils import cur_ts


class DjangoEventBridge(EventBridge):
    def receive_events(self) -> Generator[ControlEvent, None, None]:
        offset = cur_ts()
        while True:
            for event in TaskkitControlEvent.objects\
                    .filter(sent__gt=offset)\
                    .order_by('sent'):
                offset = event.sent
                if (e := decode_control_event(event.data)):
                    yield e
            time.sleep(1)

    def send_event(self, event: ControlEvent):
        TaskkitControlEvent.objects.create(
            data=encode_control_event(event),
            sent=cur_ts())


class DjangoLock(Lock):
    def __init__(self, target: str):
        self.target = target
        self.acquired = False
        self.ensured = False
        self.atomic: Optional[Atomic] = None

    def acquire(self) -> bool:
        if self.acquired:
            return True

        if not self.ensured:
            self.ensured = True
            try:
                with atomic():
                    TaskkitLock.objects\
                        .select_for_update(nowait=True)\
                        .get(pk=self.target)
            except TaskkitLock.DoesNotExist:
                try:
                    TaskkitLock.objects.create(pk=self.target)
                except IntegrityError:
                    pass
            except Exception:
                pass

        self.atomic = atomic()
        self.atomic.__enter__()
        try:
            TaskkitLock.objects\
                .select_for_update(nowait=True)\
                .get(pk=self.target)
            self.acquired = True
            return True
        except Exception:
            self.atomic.__exit__(*sys.exc_info())
            self.atomic = None
            return False

    def release(self):
        if self.atomic:
            self.atomic.__exit__(*sys.exc_info())
            self.atomic = None
            self.acquired = False


class DjangoBackend(Backend):
    def set_worker_ttl(self, worker_ids: set[str], expires_at: float):
        if not worker_ids:
            return
        workers = [
            TaskkitWorker(pk=wid, expires=expires_at)
            for wid in worker_ids
        ]
        TaskkitWorker.objects.bulk_create(workers, ignore_conflicts=True)
        TaskkitWorker.objects.bulk_update(workers, ['expires'])

    def get_workers(self) -> list[tuple[str, float]]:
        return [
            (w.pk, w.expires)
            for w in TaskkitWorker.objects.order_by('expires')
        ]

    def purge_workers(self, worker_ids: set[str]):
        if not worker_ids:
            return
        TaskkitWorker.objects.filter(pk__in=worker_ids).delete()

    def put_tasks(self, *tasks: Task):
        if not tasks:
            return
        objects = [self._task_to_db(t) for t in tasks]
        with atomic():
            self.discard_tasks(*[t.pk for t in objects])
            TaskkitTask.objects.bulk_create(objects, ignore_conflicts=True)

    def get_queued_tasks(self, group: str, limit: int) -> list[Task]:
        return [
            self._db_to_task(t) for t in TaskkitTask.objects
            .filter(began__isnull=True, group=group)
            .order_by('due')[:limit]
        ]

    def lookup_tasks(self, *task_ids: str) -> dict[str, Optional[Task]]:
        tasks = {
            t.pk: self._db_to_task(t)
            for t in TaskkitTask.objects.filter(pk__in=task_ids)
        }
        return {tid: tasks.get(tid) for tid in task_ids}

    def assign_task(self, group: str, worker_id: str) -> Optional[Task]:
        for pk in TaskkitTask.objects\
                .filter(began__isnull=True, group=group, due__lt=cur_ts())\
                .values_list('pk', flat=True)\
                .order_by('due')[:50]:
            with atomic():
                try:
                    db_task = TaskkitTask.objects\
                        .select_for_update(nowait=True)\
                        .get(pk=pk)
                    if db_task.began is None:
                        db_task.assignee_worker_id = worker_id
                        db_task.began = cur_ts()
                        db_task.save()
                        return self._db_to_task(db_task)
                except (OperationalError, TaskkitTask.DoesNotExist):
                    pass
        return None

    def discard_tasks(self, *task_ids: str):
        TaskkitTask.objects.filter(pk__in=task_ids).delete()

    def succeed(self, task: Task, result: bytes):
        with atomic():
            try:
                db_task = TaskkitTask.objects\
                    .select_for_update()\
                    .get(pk=task.id)
            except TaskkitTask.DoesNotExist:
                db_task = self._task_to_db(task)
            db_task.done = cur_ts()
            db_task.began = db_task.began or db_task.done
            db_task.result = result
            db_task.disposable = db_task.done + task.ttl
            db_task.save()

    def fail(self, task: Task, error_message: str):
        with atomic():
            try:
                db_task = TaskkitTask.objects\
                    .select_for_update()\
                    .get(pk=task.id)
            except TaskkitTask.DoesNotExist:
                db_task = self._task_to_db(task)

            db_task.done = cur_ts()
            db_task.began = db_task.began or db_task.done
            db_task.error_message = error_message
            db_task.disposable = db_task.done + task.ttl
            db_task.save()

    def get_result(self, task_id: str) -> tuple[Task, bytes]:
        try:
            db_task = TaskkitTask.objects.get(pk=task_id)
            task = self._db_to_task(db_task)
            if db_task.done is None:
                raise NoResult(task)
            if db_task.result is None:
                raise Failed(task, db_task.error_message)
            return (task, db_task.result)
        except TaskkitTask.DoesNotExist:
            raise NotFound

    def get_done_task_ids(self,
                          since: Optional[float],
                          until: Optional[float],
                          limit: int) -> list[str]:
        return list(
            TaskkitTask.objects
            .filter(done__isnull=False,
                    done__gte=since or 0,
                    done__lte=until or cur_ts())
            .values_list('pk', flat=True)
            .order_by('done')[:limit]
        )

    def get_disposable_task_ids(self, limit: int) -> list[str]:
        return list(
            TaskkitTask.objects
            .filter(disposable__isnull=False,
                    disposable__lt=cur_ts())
            .values_list('pk', flat=True)
            .order_by('disposable')[:limit]
        )

    def get_stage_info(self, limit: int) -> list[StageInfo]:
        return [
            StageInfo(t.assignee_worker_id, t.pk, t.began)
            for t in TaskkitTask.objects
            .filter(done__isnull=True, began__isnull=False)
            .only('pk', 'assignee_worker_id', 'began')
            .order_by('began')[:limit]
        ]

    def restore(self, info: StageInfo):
        try:
            task = TaskkitTask.objects\
                .filter(began__isnull=False)\
                .get(pk=info.task_id)
            task.began = None
            task.assignee_worker_id = None
            task.save()
        except TaskkitTask.DoesNotExist:
            pass

    def get_lock(self, target: str) -> Lock:
        return DjangoLock(target)

    def persist_scheduler_state_and_put_tasks(self,
                                              name: str,
                                              data: bytes,
                                              *tasks: Task):
        with atomic():
            try:
                state = TaskkitSchedulerState.objects.get(id=name)
                state.data = data
                state.save()
            except TaskkitSchedulerState.DoesNotExist:
                TaskkitSchedulerState.objects.create(id=name, data=data)
            self.put_tasks(*tasks)

    def get_scheduler_state(self, name: str) -> Optional[bytes]:
        return TaskkitSchedulerState.objects\
            .filter(pk=name)\
            .values_list('data', flat=True)\
            .first()

    def destroy_all(self):
        with atomic():
            TaskkitLock.objects.all().delete()
            TaskkitWorker.objects.all().delete()
            TaskkitTask.objects.all().delete()
            TaskkitSchedulerState.objects.all().delete()

    def _db_to_task(self, t: TaskkitTask) -> Task:
        return Task(
            id=t.pk,
            name=t.name,
            data=t.data,
            due=t.due,
            created=t.created,
            scheduled=t.scheduled,
            retry_count=t.retry_count,
            ttl=t.ttl,
        )

    def _task_to_db(self, t: Task) -> TaskkitTask:
        return TaskkitTask(
            id=t.id,
            group=t.group,
            name=t.name,
            data=t.data,
            due=t.due,
            created=t.created,
            scheduled=t.scheduled,
            retry_count=t.retry_count,
            ttl=t.ttl,
            assignee_worker_id=None,
            began=None,
            result=None,
            done=None,
            disposable=None,
        )


def make_kit(handler: TaskHandler) -> Kit:
    return Kit(DjangoBackend(), DjangoEventBridge(), handler)



from .backend import Lock, NotFound, NoResult, Failed, Backend  # noqa
from .event import ControlEvent, EventBridge  # noqa
from .kit import Kit, ScheduleEntryDict, ScheduleEntryCompat, ScheduleEntriesCompat, ScheduleEntriesCompatMapping  # noqa
from .process import TaskkitProcess  # noqa
from .result import Result, ResultGetTimedOut  # noqa
from .scheduler import Schedule, DuplicationPolicy, OnlyEarliest, OnlyLatest, RegularSchedule, ScheduleEntry  # noqa
from .stage import StageInfo   # noqa
from .task import Task, DiscardTask, TaskHandler  # noqa
from .version import VERSION  # noqa



from dataclasses import dataclass, fields
from datetime import datetime, timezone
from typing import Any, Protocol, Optional, Union
from uuid import uuid4

from .utils import cur_ts, as_ts, from_ts, local_tz


DEFAULT_TASK_TTL = 60 * 60 * 24 * 7


@dataclass(frozen=True)
class Task:
    id: str
    name: str
    data: bytes
    due: float
    created: float
    scheduled: Optional[float]
    retry_count: int
    ttl: float

    @classmethod
    def init(cls,
             group: str,
             name: str,
             data: bytes,
             due: Union[datetime, float, None] = None,
             scheduled: Union[datetime, float, None] = None,
             ttl: float = DEFAULT_TASK_TTL) -> 'Task':
        created = cur_ts()
        return cls(id=cls.make_id(group),
                   name=name,
                   data=data,
                   due=as_ts(due or created),
                   created=created,
                   scheduled=as_ts(scheduled),
                   retry_count=0,
                   ttl=ttl)

    @property
    def group(self) -> str:
        return self.get_group_from_id(self.id)

    def due_dt(self, tzinfo: Optional[timezone] = None) -> datetime:
        return from_ts(self.due, tzinfo or local_tz())

    def created_dt(self, tzinfo: Optional[timezone] = None) -> datetime:
        return from_ts(self.created, tzinfo or local_tz())

    def scheduled_dt(self, tzinfo: Optional[timezone] = None) -> Optional[datetime]:
        return from_ts(self.scheduled, tzinfo or local_tz())\
            if self.scheduled is not None else None

    @staticmethod
    def make_id(group: str) -> str:
        return f'{group}_{uuid4().hex}'

    @staticmethod
    def get_group_from_id(task_id: str) -> str:
        return task_id.rsplit('_', 1)[0]

    @staticmethod
    def group_tasks(*tasks: 'Task') -> dict[str, list['Task']]:
        result: dict[str, list[Task]] = dict()
        for task in tasks:
            group = task.group
            if group not in result:
                result[group] = [task]
            else:
                result[group].append(task)
        return result

    @staticmethod
    def group_task_ids(*task_ids: str) -> dict[str, list[str]]:
        result: dict[str, list[str]] = dict()
        for tid in task_ids:
            group = Task.get_group_from_id(tid)
            if group not in result:
                result[group] = [tid]
            else:
                result[group].append(tid)
        return result

    def to_dict(self) -> dict[str, Any]:
        return {
            f.name: getattr(self, f.name)
            for f in fields(self)
        }

    def clone_for_retry(self, interval: float) -> 'Task':
        d = self.to_dict()
        d['due'] = cur_ts() + interval
        d['retry_count'] += 1
        return Task(**d)


class DiscardTask(Exception):
    pass


class TaskHandler(Protocol):
    def handle(self, task: Task) -> Any:
        """Handle the task or raise DiscardTask to ignore the task"""
        ...

    def get_retry_interval(self,
                           task: Task,
                           exception: Exception) -> Optional[float]:
        """Return interval in seconds if you want to retry the failed task.

        This method will be called if the handle method raises exceptions. You
        should return how long time should be wait to retry the task in seconds
        as float. If you don't want to retry the task, you can return None to
        make the task fail or raise DiscardTask to discard the task.
        """
        ...

    def encode_data(self, group: str, task_name: str, data: Any) -> bytes:
        ...

    def encode_result(self, task: Task, result: Any) -> bytes:
        ...

    def decode_result(self, task: Task, encoded: bytes) -> Any:
        ...



# Generated by Django 4.1.3 on 2022-11-09 06:43

from django.db import migrations, models


class Migration(migrations.Migration):

    initial = True

    dependencies = [
    ]

    operations = [
        migrations.CreateModel(
            name='TaskkitControlEvent',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sent', models.FloatField(db_index=True)),
                ('data', models.BinaryField()),
            ],
            options={
                'db_table': 'taskkit_control_event',
            },
        ),
        migrations.CreateModel(
            name='TaskkitLock',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
            ],
            options={
                'db_table': 'taskkit_lock',
            },
        ),
        migrations.CreateModel(
            name='TaskkitSchedulerState',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('data', models.BinaryField()),
            ],
            options={
                'db_table': 'taskkit_scheduler_state',
            },
        ),
        migrations.CreateModel(
            name='TaskkitWorker',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('expires', models.FloatField(db_index=True)),
            ],
            options={
                'db_table': 'taskkit_worker',
            },
        ),
        migrations.CreateModel(
            name='TaskkitTask',
            fields=[
                ('id', models.CharField(max_length=255, primary_key=True, serialize=False)),
                ('group', models.CharField(max_length=40)),
                ('name', models.CharField(max_length=255)),
                ('data', models.BinaryField()),
                ('due', models.FloatField()),
                ('created', models.FloatField()),
                ('scheduled', models.FloatField(null=True)),
                ('retry_count', models.PositiveIntegerField()),
                ('ttl', models.FloatField()),
                ('assignee_worker_id', models.CharField(max_length=255, null=True)),
                ('began', models.FloatField(null=True)),
                ('result', models.BinaryField(null=True)),
                ('error_message', models.TextField(null=True)),
                ('done', models.FloatField(null=True)),
                ('disposable', models.FloatField(db_index=True, null=True)),
            ],
            options={
                'db_table': 'taskkit_task',
                'index_together': {('done', 'began'), ('name', 'created'), ('began', 'group', 'due')},
            },
        ),
    ]



import os
from datetime import tzinfo
from multiprocessing import Process, Event

from .backend import Backend
from .event import EventBridge
from .scheduler import ScheduleEntry, Scheduler
from .signal import SignalCaptured, capture_signals, signal
from .task import TaskHandler
from .threads import WorkerThread, ServiceThread
from .services import RefreshWorkerLifetime, PurgeDeadWorkers,\
    RestoreAbandonedTasks, DiscardDisposableTasks
from .utils import logger


class TaskkitProcess(Process):
    def __init__(self,
                 num_worker_threads_per_group: dict[str, int],
                 backend: Backend,
                 bridge: EventBridge,
                 handler: TaskHandler,
                 schedule_entries: dict[str, list[ScheduleEntry]],
                 tzinfo: tzinfo,
                 **kwargs):
        assert all(n >= 0 for n in num_worker_threads_per_group.values()),\
            'All values for num_worker_threads_per_group must be positive int.'
        self.num_worker_threads = num_worker_threads_per_group
        self.backend = backend
        self.bridge = bridge
        self.handler = handler
        self.schedule_entries = schedule_entries
        self.tzinfo = tzinfo
        self._terminate_event = Event()
        super().__init__(**kwargs)

    def is_active(self) -> bool:
        return not self._terminate_event.is_set()

    def run(self):
        _id = f'{os.getpid()}'
        logger.info(f'[{_id}] taskkit process started: '
                    f'{self.num_worker_threads}')

        backend = self.backend
        workers: list[WorkerThread] = []
        for group, n in self.num_worker_threads.items():
            for _ in range(n):
                w = WorkerThread(group, backend, self.handler)
                w.start()
                workers.append(w)
        refresh_ttl = ServiceThread(
            RefreshWorkerLifetime(backend, workers))
        refresh_ttl.start()

        services = [
            *[ServiceThread(Scheduler(name, backend, entries, self.tzinfo))
              for name, entries in self.schedule_entries.items() if entries],
            ServiceThread(RestoreAbandonedTasks(backend)),
            ServiceThread(PurgeDeadWorkers(backend)),
            ServiceThread(DiscardDisposableTasks(backend)),
        ]
        for service in services:
            service.start()

        alive = {g for g, n in self.num_worker_threads.items() if n > 0}

        try:
            with capture_signals(signal.SIGTERM):
                while True:
                    try:
                        for event in self.bridge.receive_events():
                            target_groups =\
                                (alive
                                 if (_groups := event['groups']) is None
                                 else _groups)
                            target_workers = [w for w in workers
                                              if w.group in target_groups]

                            logger.info(f'[{_id}] control event received: {event}')

                            if event['name'] == 'shutdown':
                                alive.difference_update(target_groups)
                                for worker in target_workers:
                                    worker.should_stop = True
                                if not alive:
                                    raise SystemExit
                            elif event['name'] == 'pause':
                                for worker in target_workers:
                                    worker.paused = True
                            elif event['name'] == 'resume':
                                for worker in target_workers:
                                    worker.paused = False
                    except Exception:
                        pass
        except (KeyboardInterrupt, SystemExit, SignalCaptured):
            self._terminate_event.set()

        logger.info(f'[{_id}] shutting down...')

        for w in workers:
            w.should_stop = True
        for service in services:
            service.should_stop = True
        for service in services:
            service.join()
        for w in workers:
            w.join()
        refresh_ttl.should_stop = True
        refresh_ttl.join()



import time
from threading import Thread
from .backend import Backend
from .services import Service
from .task import TaskHandler
from .utils import logger
from .worker import Worker


class WorkerThread(Thread):
    def __init__(self,
                 group: str,
                 backend: Backend,
                 handler: TaskHandler,
                 **kwargs):
        self.group = group
        self.worker = Worker(group, backend, handler)
        self.should_stop = False
        self.paused = False
        super().__init__(**kwargs)

    @property
    def id(self) -> str:
        return self.worker.id

    def get_id(self) -> str:
        return self.worker.id

    def run(self):
        logger.info(f'[{self.id}] worker thread started')
        while not self.should_stop:
            if self.paused:
                time.sleep(1)
                continue

            if not self.worker():
                time.sleep(1)
        logger.info(f'[{self.id}] worker thread stopped')


class ServiceThread(Thread):
    def __init__(self,
                 service: Service,
                 **kwargs):
        self.should_stop = False
        self.service = service
        super().__init__(**kwargs)

    def run(self):
        while not self.should_stop:
            try:
                time_to_wait = self.service()
                while time_to_wait > 0:
                    time.sleep(min(1, time_to_wait))
                    time_to_wait -= 1
                    if self.should_stop:
                        return
            except Exception:
                logger.exception(
                    'Unexpected exception occurred during service call '
                    f'({self.service.__class__.__qualname__})')
                time.sleep(1)
